/*
 * connection.cpp
 *
 * Kai G. Schwebke
 * 2007-03-03
 *
 * Client connection handling.
 *
 */
#include "connection.h"
#include <cstdio>
#include <cstdlib>
#include <string>
#include <vector>
#include <unistd.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/types.h>
#include "config.h"
#include "logging.h"
#include "sensormanager.h"
#include "sensor.h"
#include "tool.h"

using namespace std;

Connection::Connection(ConnectionManager* manager, int fd) {
   this->manager = manager;
   this->fd = fd;
   pthread_attr_t attr;
   pthread_attr_init(&attr);
   pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
   if (pthread_create(&thread, &attr, (void*(*)(void*))connectionThread, this)) {
      logEvent(LOG_CRITICAL, "failed to spawn connection thread");
   }
   pthread_attr_destroy(&attr);
}

Connection::~Connection() {
   close(fd);
}

void Connection::startServing() {
   // prepare select file descriptor set
   fd_set inputFds;
   FD_ZERO(&inputFds);
   FD_SET(fd, &inputFds);

   // input data buffer
   string inBuff("");

   // serve the connection until other side disconnects or daemon should shutdown
   while (!shouldShutdown) {
      struct timeval timeout;
      timeout.tv_sec  = CHECK_INTERVAL_SEC;
      timeout.tv_usec = CHECK_INTERVAL_USEC;

      // copy file descriptor set
      fd_set fds = inputFds;

      // wait for connection or check interval timeout
      if (select(fd+1, &fds, 0, 0, &timeout) < 0) {
	 // ignore error caused by signal
	 if ( errno != EINTR ) {
	    // close connection on other errors
	    manager->closeConnection(this);
	    return;
	 }
      } else {
	 // no error
	 // process incoming connection, if any
	 // do nothing on timeout
	 if (FD_ISSET(fd, &fds)) {
	    char recBuff[1024];
	    int r = recv(fd, &recBuff, 1024, 0);

	    if (r <= 0) {
	       // close connection on EOF or error
	       logEvent(LOG_STAT, "lost client connection");
	       manager->closeConnection(this);
	       return;
	    } else {
	       // append received data to input buffer
	       inBuff.append(recBuff, r);

	       // while buffer contains complete lines
	       string::size_type pos;
	       while ( (pos = inBuff.find_first_of("\n\r", 0)) != string::npos ) {
		  // process line
		  string line(inBuff.substr(0, pos));
		  if (process(line)) {
		     logEvent(LOG_STAT, "error sending client reply");
		     manager->closeConnection(this);
		     return;
		  }

		  // remove line from buffer
		  string::size_type lpos =
		     inBuff.find_first_not_of("\n\r", pos);
		  if (lpos != string::npos) {
		     inBuff.erase(0, lpos);
		  } else {
		     inBuff.erase();
		  }
	       }
	    }
	 }
      }
      // loop until server should shutdown
   }
}

int Connection::process(string line) {
   string message("received '");
   message += line;
   message += "'";
   logEvent(LOG_DEBUG, message);

   string reading = manager->getSensorManager()->getReading(line);

   message = "sending '";
   message += reading;
   message += "'";
   logEvent(LOG_DEBUG, message);

   reading += REPLY_NEWLINE;
   size_t readingSize = reading.length();
   return sendall(fd, reading.c_str(), &readingSize);
}

void* connectionThread(Connection* c) {
   c->startServing();
}
