/*
 * connectionmanager.cpp
 *
 * Kai G. Schwebke
 * 2007-03-02
 *
 * Listening for incoming connections.
 * Creates and manages Connection objects.
 *
 */
#include "connectionmanager.h"
#include <cstdio>
#include <cstdlib>
#include <string>
#include <unistd.h>
#include <pthread.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include "config.h"
#include "tool.h"
#include "logging.h"

using namespace std;

ConnectionManager::ConnectionManager(SensorManager* sensorManager) {
   this->sensorManager = sensorManager;
   pthread_mutex_init(&mutex, 0);
}

ConnectionManager::~ConnectionManager() {
   pthread_mutex_lock(&mutex);

   // delete connections
   list<Connection*>::iterator i;
   for (i = connections.begin(); i != connections.end(); ++i) {
      delete *i;
   }

   // close listen socket
   if (listenSocketFd >= 0) {
      close(listenSocketFd);
   }

   pthread_mutex_unlock(&mutex);
   pthread_mutex_destroy(&mutex);
}

void ConnectionManager::startListen(int listenPort) {
   // create socket
   listenSocketFd = socket(PF_INET, SOCK_STREAM, 0);
   if (listenSocketFd < 0) {
      logEvent(LOG_CRITICAL, "failed to create socket");
      return;
   }
   int yes = 1;
   if (setsockopt(listenSocketFd, SOL_SOCKET, SO_REUSEADDR,
	          &yes, sizeof(int)) < 0) {
      logEvent(LOG_CRITICAL, "failed to set socket options");
      return;
   }
   
   // listen on port (bind to any-address)
   struct sockaddr_in local_addr;
   local_addr.sin_family = AF_INET;
   local_addr.sin_port = htons(listenPort);
   local_addr.sin_addr.s_addr = INADDR_ANY;
   memset(&(local_addr.sin_zero), '\0', 8);
   if (bind(listenSocketFd, (struct sockaddr*)&local_addr, sizeof(struct sockaddr)) == -1) {
      string message("failed to bind socket (");
      message += strerror(errno);
      message += ")";
      logEvent(LOG_CRITICAL, message);
      return;
   }
   if (listen(listenSocketFd, LISTEN_PORT_BACKLOG) == -1) {
      string message("failed to listen socket (");
      message += strerror(errno);
      message += ")";
      logEvent(LOG_CRITICAL, message);
      return;
   }
   string message("waiting for incoming connections on port ");
   message += itoa(listenPort);
   logEvent(LOG_STAT, message);

   // prepare select file descriptor set
   fd_set listenSocketFds;
   FD_ZERO(&listenSocketFds);
   FD_SET(listenSocketFd, &listenSocketFds);

   // accept incoming connections until daemon should shutdown
   while (!shouldShutdown) {
      struct timeval timeout;
      timeout.tv_sec  = CHECK_INTERVAL_SEC;
      timeout.tv_usec = CHECK_INTERVAL_USEC;

      // copy file descriptor set
      fd_set fds = listenSocketFds;

      // wait for connection or check interval timeout
      if (select(listenSocketFd+1, &fds, 0, 0, &timeout) < 0) {
	 // ignore error caused by signal
	 if ( errno != EINTR ) {
	    // abort on other errors
	    logEvent(LOG_CRITICAL, "select on listen socket failed");
	    return;
	 }
      } else {
	 // no error
	 // process incoming connection, if any
	 // do nothing on timeout
	 if (FD_ISSET(listenSocketFd, &fds)) {
	    struct sockaddr_in remote_addr;
	    socklen_t sin_size = sizeof(struct sockaddr_in);
	    int new_fd = accept(listenSocketFd, (struct sockaddr*)&remote_addr, &sin_size);
	    if (new_fd < 0) {
	       string message("failed to accept connection (");
	       message += strerror(errno);
	       message += ")";
	       logEvent(LOG_CRITICAL, message);
	       return;
	    }
	    string message("incoming connection from ");
	    message += inet_ntoa(remote_addr.sin_addr);
	    logEvent(LOG_STAT, message);

	    newConnection(new_fd);
	 }
      }
      // loop until server should shutdown
   }

}

void ConnectionManager::newConnection(int fd) {
   pthread_mutex_lock(&mutex);
   Connection* c = new Connection(this, fd);
   connections.push_back(c);
   pthread_mutex_unlock(&mutex);
}

void ConnectionManager::closeConnection(Connection* c) {
   pthread_mutex_lock(&mutex);
   connections.remove(c);
   delete c;
   pthread_mutex_unlock(&mutex);
}
