/*
 * main.cpp
 *
 * Kai G. Schwebke
 * 2007-03-02
 *
 * Main program.
 *
 */
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <string>
#include <unistd.h>
#include <signal.h>
#include "tool.h"
#include "ciniparser.h"
#include "config.h"
#include "logging.h"
#include "daemon.h"
#include "sighandler.h"
#include "sensormanager.h"
#include "connectionmanager.h"

using namespace std;


// manager object for sensors
SensorManager sensorManager;

// manager object for connections
ConnectionManager connectionManager(&sensorManager);


int main(int argc, char* argv[]) {
   printf("pcmeasureUSB %s\n\n", VERSION);

   // flag indicating if program should become a daemon
   bool daemon = false;

   // initialize logging
   logInit();


   //////////////////////////////////////////////////////////////////
   // parse command line
   while (1) {
      int c = getopt(argc, argv, "?hd");
      if (c == -1) break;
      switch (c) {
	 case '?':
	    showHelp();
	    return 1;

	 case 'h':
	    showHelp();
	    return 0;

	 case 'd':
	    daemon = true;
	    break;
      }
   }

   const char* configFile = DEFAULT_CONFIG_FILE;
   if (optind < argc) {
      configFile = argv[optind];
   }

   if ((optind+1) < argc) {
      printf("trailing arguments on command line\n");
      showHelp();
      return 1;
   }


   //////////////////////////////////////////////////////////////////
   // parse config file
   //    preset defaults
   //       set default log file in daemon mode
   if (daemon) {
      logFile = DEFAULT_LOG_FILE;
   }
   //       set default listen port
   int listenPort = DEFAULT_LISTEN_PORT;

   //    iniparser dictionary
   dictionary* d = 0;

   //    try to load ini file from current directory
   d = iniparser_load(configFile);

   //    if not found, try to load from central location
   if (d == 0) {
      string configFile(DEFAULT_CONFIG_PATH);
      configFile += "/";
      configFile += DEFAULT_CONFIG_FILE;
      d = iniparser_load(configFile.c_str());

      if ( (d != 0) && (!daemon) ) {
	 // report used config file if not in daemon mode
	 string message("using config file '");
	 message += configFile;
	 message += "'";
	 logEvent(LOG_STAT, message);
      }
   } else {
      if (!daemon) {
	 // report used config file if not in daemon mode
	 string message("using config file './");
	 message += configFile;
	 message += "'";
	 logEvent(LOG_STAT, message);
      }
   }
   if (d == 0) {
      // no config file, use built in defaults
      if (!daemon)
	 logEvent(LOG_STAT, "no config file found, using defaults");
   } else {
      // config file found
      //    parse port (if present, otherwise use default)
      int configPort = iniparser_getint(d, "main:port", 0);
      if (configPort > 0) {
	 listenPort = configPort;
      }

      //    parse log file (if present, otherwise use default)
      const char* configLogFile = iniparser_getstring(d, "main:logfile", 0);

      //    override default log file if configured and in daemon mode
      if (configLogFile && daemon){
	 logFile = configLogFile;
      }

      //    parse log level
      const char* configLogLevel = iniparser_getstring(d, "main:loglevel", 0);
      if (configLogLevel) {
	 if (!strcmp(configLogLevel, "critical")) {
	    setLogLevel(LOG_CRITICAL);
	 } else if (!strcmp(configLogLevel, "stat")) {
	    setLogLevel(LOG_STAT);
	 } else if (!strcmp(configLogLevel, "debug")) {
	    setLogLevel(LOG_DEBUG);
	 } else {
	    logEvent(LOG_CRITICAL, "configured log level unknown");
	 }
      }

      //    parse sensors (if present, otherwise defaults will be added later)
      int numSensors = iniparser_getnseckeys(d, "sensors");
      int i;
      for (i = 0; i < numSensors; ++i) {
	 // fetch sensor device and assigned name(s) from config
	 const char* key = iniparser_getseckeyname(d, "sensors", i);
	 const char* sensorDevice = key + strlen("sensors") + 1;
	 string val(iniparser_getstring(d, key, 0));
	 vector<string> sensorNames;
	 stringTokenize(val, sensorNames);
	 if (val.size() == 0) {
	    string message("no names specified for sensor '");
	    message += sensorDevice;
	    message += "'";
	    logEvent(LOG_CRITICAL, message);
	    shutdown();
	 }

	 // create and name sensor
	 string message("adding sensor '");
	 message += sensorDevice;
	 message += "'";
	 logEvent(LOG_STAT, message);
	 Sensor* sensor = sensorManager.newSensor(sensorDevice);
	 for (vector<string>::iterator i = sensorNames.begin();
	      i != sensorNames.end();
	      ++i) {
	    string message("   --> naming sensor '");
	    message += *i;
	    message += "'";
	    logEvent(LOG_STAT, message);
	    sensorManager.nameSensor(sensor, *i);
	 }
      }

      iniparser_freedict(d);
   }

   // add default sensors if none are configured
   if (sensorManager.sensorCount() == 0) {
      Sensor* defaultSensor = sensorManager.newSensor(DEFAULT_SENSOR_DEVICE);
      sensorManager.nameSensor(defaultSensor, DEFAULT_SENSOR_NAME);
   }


   //////////////////////////////////////////////////////////////////
   // daemonize process if requested
   if (daemon) {
      logEvent(LOG_CRITICAL, "starting pcmeasureUSB as daemon");
      daemonize();
   } else {
      logEvent(LOG_CRITICAL, "starting pcmeasureUSB");
   }

   
   //////////////////////////////////////////////////////////////////
   // install signal handlers for interrupt and terminate
   // (these handlers flag shouldShutdown)
   signal(SIGINT, catch_shutdown_signal);
   signal(SIGTERM, catch_shutdown_signal);


   //////////////////////////////////////////////////////////////////
   // start sensor reading threads
   sensorManager.createThreads();


   //////////////////////////////////////////////////////////////////
   // start listening for inbound connections until shutdown
   connectionManager.startListen(listenPort);

   shutdown();

   // never reached
   return 0;
}


