/*
 * sensor.cpp
 *
 * Kai G. Schwebke
 * 2007-03-03
 *
 * Sensor data handling.
 *
 */
#include "sensor.h"
#include <errno.h>
#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include "sensormanager.h"
#include "logging.h"
#include "serial.h"
#include "config.h"
#include "tool.h"

using namespace std;

Sensor::Sensor(SensorManager* manager, string deviceName) {
   this->manager = manager;
   this->deviceName = deviceName;
   readTime = 0;
   for (int i = 1; i <= ReadNameCount; ++i) {
      readContent[ReadName[i]] = "";
   }
   pthread_mutex_init(&mutex, 0);
}

Sensor::~Sensor() {
   pthread_mutex_destroy(&mutex);
}

void Sensor::createThread() {
   pthread_attr_t attr;
   pthread_attr_init(&attr);
   pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
   if (pthread_create(&thread, 0, (void*(*)(void*))sensorThread, this)) {
      logEvent(LOG_CRITICAL, "failed to spawn sensor thread");
   }
   pthread_attr_destroy(&attr);
}

void Sensor::startSensing() {
   string message("start sensing on '");
   message += deviceName;
   message += "'";
   logEvent(LOG_STAT, message);
   Serial serial(deviceName);

   int fd = serial.getFd();
   if (fd >= 0) {
      // prepare select file descriptor set
      fd_set inputFds;
      FD_ZERO(&inputFds);
      FD_SET(fd, &inputFds);

      // input data buffer
      string inBuff("");

      // serve the sensor until other side disconnects or daemon should shutdown
      while (!shouldShutdown) {
	 struct timeval timeout;
	 timeout.tv_sec  = CHECK_INTERVAL_SEC;
	 timeout.tv_usec = CHECK_INTERVAL_USEC;

	 // copy file descriptor set
	 fd_set fds = inputFds;

	 // wait for connection or check interval timeout
	 if (select(fd+1, &fds, 0, 0, &timeout) < 0) {
	    // ignore error caused by signal
	    if ( errno != EINTR ) {
	       // stop sensing on other errors
	       return;
	    }
	 } else {
	    // no error
	    // process incoming connection, if any
	    // do nothing on timeout
	    if (FD_ISSET(fd, &fds)) {
	       char recBuff[1024];
	       int r = read(fd, &recBuff, 1024);

	       if (r <= 0) {
		  // close connection on EOF or error
		  return;
	       } else {
		  // append received data to input buffer
		  inBuff.append(recBuff, r);

		  // while buffer contains complete lines
		  string::size_type pos;
		  while ( (pos = inBuff.find_first_of("\n\r", 0)) != string::npos ) {
		     // process line
		     if (pos > 0) {
			string line(inBuff.substr(0, pos));
			process(line);
		     }

		     // remove line from buffer
		     string::size_type lpos =
			inBuff.find_first_not_of("\n\r", pos);
		     if (lpos != string::npos) {
			inBuff.erase(0, lpos);
		     } else {
			inBuff.erase();
		     }
		  }
	       }
	    }
	 }
	 // loop until server should shutdown
      }
   }
}

void Sensor::process(string line) {
   string message("sensed '");
   message += line;
   message += "' on '";
   message += deviceName;
   message += "'";
   logEvent(LOG_DEBUG, message);

   // split input on ' '
   vector<string> tokens;
   stringTokenize(line, tokens);

   pthread_mutex_lock(&mutex);

   for (vector<string>::iterator i = tokens.begin();
	i != tokens.end();
	++i) {
      // split value on '='
      vector<string> values;
      stringTokenize(*i, values, "=");
      if (values.size() == 2) {
	 // strip unit, if present
	 string::size_type pos = values[1].find_first_not_of("01234567890.");
	 if (pos == string::npos)
	 {
	    readContent[values[0]] = values[1];
	 } else {
	    readContent[values[0]] = values[1].substr(0, pos);
	 }

	 // update time of reading
	 readTime = time(0);
      }
   }

   pthread_mutex_unlock(&mutex);
}

string Sensor::getReading(ReadType readType) {
   string reading("valid=0");

   pthread_mutex_lock(&mutex);

   // return valid=0 if last sensing is too old
   if (time(0) > (readTime+SENSOR_READING_LIFETIME_SEC)) {
      pthread_mutex_unlock(&mutex);
      return reading;
   }

   // return last sensing
   if ((readType > 0) && (readType <= ReadNameCount)) {
      reading = "valid=1;value=";
      reading += readContent[ReadName[readType]];
      reading += ";";
   }

   pthread_mutex_unlock(&mutex);
   return reading;
}

const int Sensor::ReadNameCount = 3;
const char * const Sensor::ReadName[] = {
   "",
   "temperature",
   "humidity",
   "dewpoint"};

void* sensorThread(Sensor* s) {
   while (1) {
      s->startSensing();

      if (shouldShutdown) {
	 return 0;
      }

      // Connection to sensor lost.
      // Wait reconnect retry time, then reconnect.
      // During wait, every check interval the shutdown
      // flag is checked.
      string message("lost sensor connection on '");
      message += s->getDeviceName();
      message += "'";
      logEvent(LOG_STAT, message);

      time_t start = time(0);
      while ( time(0) < (start+SENSOR_RECONNECT_SEC) ) {
	 if (CHECK_INTERVAL_SEC > 0) {
	    sleep(CHECK_INTERVAL_SEC);
	 }
	 if (CHECK_INTERVAL_USEC > 0) {
	    usleep(CHECK_INTERVAL_USEC);
	 }
	 if (shouldShutdown) {
	    return 0;
	 }
      }
   }
}
